/*
	
					The Hacker's Diet

					  Trend Fitting
					  
	This file contains code which fits a linear trend to a
	collection of day to day trend values using linear regression
	with the least squares method.  Two versions of the code
	are implemented.  The first uses fixed-point integer
	arithmetic, which is adequate for trend computations spanning
	at least a year and is much faster than floating point.  The
	second version uses floating point, and is employed when
	the need arises to fit a trend over a period sufficiently
	long that the fixed-point code would overflow.  The fixed
	point code contains assertions to detect overflow, should it
	occur.
	
	The low-level accumulation and analysis functions are
	externalised for the benefit of (as of this writing)
	History.c, to enable it to fit a trend at the same time
	it is plotting the trend on the chart.  This avoids a
	second costly traversal of the database with the overhead
	of unpacking the month and recomputing the trend.
					 
*/

#include "Hackdiet.h"

#define TrendToDouble(t) ((t) / 1000000.0)

//	Fixed point internal functions

#define HYB_SCALE	1000

//	fitSlopeStart  --  Initialise trend fit

void fitSlopeStart(struct iTrendFitContext *c)
{
	c->n = 0;
	c->s1 = 0;
	c->s2 = c->s3 = c->s4 = 0;
}

//	fitSlopePoint  --  Add a point on the trend curve to the calculation

void fitSlopePoint(struct iTrendFitContext *c, trend t)
{
	c->s1 += (c->n + 1) * TrendToDouble(t);
	c->s2 += (c->n + 1);
	assert(c->s2 >= 0);
	c->s3 += (t / HYB_SCALE);
	assert(c->s3 >= 0);
	c->s4 += ((long) (c->n + 1)) * (c->n + 1);
	assert(c->s4 >= 0);
	c->n++;
}

//	fitSlopeEnd  --  Complete calculation and return best fit slope

double fitSlopeEnd(struct iTrendFitContext *c)
{
	double ds2 = c->s2,
		   ds4 = ((double) c->s4) * c->n;

	c->s1 *= c->n;
	return (c->s1 - (ds2 * (c->s3 / (1000000.0 / HYB_SCALE)))) / (ds4 - (ds2 * ds2));
}

//	Floating point internal functions

static int n = 0;
static double ds1, ds2, ds3, ds4;		// Terms and factors in linear regression relation

//	DfitSlopeStart  --  Initialise trend fit

void DfitSlopeStart(void)
{
	assert(n == 0);
	n = 0;
	ds1 = ds2 = ds3 = ds4 = 0;
}

//	DfitSlopePoint  --  Add a point on the trend curve to the calculation

void DfitSlopePoint(trend t)
{
	double td = TrendToDouble(t);
	
	ds1 += (n + 1) * td;
	ds2 += (n + 1);
	ds3 += td;
	ds4 += ((double) (n + 1)) * (n + 1);
	n++;
}

//	DfitSlopeEnd  --  Complete calculation and return best fit slope

double DfitSlopeEnd(void)
{
	ds1 *= n;
	ds4 *= n;
	
	n = 0;
	return (ds1 - (ds2 * ds3)) / (ds4 - (ds2 * ds2));
}

/*	fitSlopeToTrend  --  Linear regression fit slope to trend for a
						 range of days within a month.  Calls the
						 more general routines above which are capable
						 of spanning month boundaries.  Since the fixed
						 point code cannot possibly overflow when fitting
						 within a month, it is always used.  The result is
						 returned in the weightUnit (log unit) specified
						 in the preferences, regardless of the unit in
						 which the monthly log happens to be stored.  */

double fitSlopeToTrend(struct hdMonth *m, UInt16 firstDay, UInt16 lastDay)
{
	int i, n;
	struct iTrendFitContext ctx;
	
	assert(firstDay <= lastDay);
	assert(lastDay < m->days);
	
	n = (lastDay - firstDay) + 1;
	
	fitSlopeStart(&ctx);
	
	for (i = 0; i < n; i++) {
		fitSlopePoint(&ctx, cvtTrend(m->trendArray[firstDay + i], (weightunit) m->weightUnit, prefs.weightUnit));
	}
	
	return fitSlopeEnd(&ctx);
}

/*	fitSlopeToMultiMonthTrend  --  Fit trend to days in the given
								   range of dates.  If the start date
								   or end dates are not present in the
								   database, an error is signified by
								   returning false.  Otherwise the slope
								   of the trend is stored into the slope
								   argument, compensating if necessary for
								   missing months between the start and end.
								   It is permissible to call this function
								   with the same year and month for start
								   and end, but if you know the range is
								   within a single month it's faster to use
								   fitSlopeToTrend above, especially if the
								   unpacked month is alread in memory.
								   The result is returned in the weightUnit
								   (log unit) specified in the preferences,
								   regardless of the unit in which the monthly
								   logs happen to be stored.  */
								   
Boolean fitSlopeToMultiMonthTrend(UInt16 sYear, UInt16 sMonth, UInt16 sDay,
								  UInt16 eYear, UInt16 eMonth, UInt16 eDay,
								  double *slope)
{
	struct hdMonth *sm = NULL, *em = NULL;
	UInt16 i, sday, recnr;
	Boolean ok = false;
	trend ltrend;
	struct iTrendFitContext ctx;
	
	sm = hdGetMonthByDate(sYear, sMonth, &recnr, false);
	if ((sYear == eYear) && (sMonth == eMonth)) {
		em = sm;
	} else {
		em = hdGetMonthByDate(eYear, eMonth, &recnr, false);
	}
	
	if ((sm != NULL) && (em != NULL)) {
		fitSlopeStart(&ctx);
		sday = sDay;
		while (sm != em) {
			if (sm != NULL) {
				for (i = sday; i < sm->days; i++) {
					fitSlopePoint(&ctx, cvtTrend(sm->trendArray[i], (weightunit) sm->weightUnit, prefs.weightUnit));
				}
				ltrend = cvtTrend(sm->trendArray[sm->days - 1], (weightunit) sm->weightUnit, prefs.weightUnit);
				hdDisposeMonthUnpacked(sm);
			} else {
				//	Month within range not in database; fake it
				
				for (i = sday; i < DaysInMonth(sMonth + 1, sYear); i++) {
					fitSlopePoint(&ctx, ltrend);
				}
			}
			sday = 0;
			sMonth++;
			if (sMonth >= 12) {
				sYear++;
				sMonth = 0;
			}
			if ((sYear == eYear) && (sMonth == eMonth)) {
				sm = em;
			} else {
				sm = hdGetMonthByDate(sYear, sMonth, &recnr, false);
			}
		}
		for (i = sday; i <= eDay; i++) {
			fitSlopePoint(&ctx, cvtTrend(em->trendArray[i], (weightunit) em->weightUnit, prefs.weightUnit));
		}
		*slope = fitSlopeEnd(&ctx);
		ok = true;
	}
	
	if (sm != NULL) {
		if (sm == em) {
			em = NULL;
		}
		hdDisposeMonthUnpacked(sm);
	}
	if (em != NULL) {
		hdDisposeMonthUnpacked(em);
	}
	return ok;
}

#ifdef NEED_DFIT_MULTI

/*	DfitSlopeToMultiMonthTrend  --  Fit trend to days in the given
									range, using floating point arithmetic.
									Use this when you need to determine a trend
									over a period sufficiently long that the
									fixed point computation used by
									fitSlopeToMultiMonthTrend() above
									might overflow.  */

								   
Boolean DfitSlopeToMultiMonthTrend(UInt16 sYear, UInt16 sMonth, UInt16 sDay,
								   UInt16 eYear, UInt16 eMonth, UInt16 eDay,
								   double *slope)
{
	struct hdMonth *sm = NULL, *em = NULL;
	UInt16 i, sday, recnr;
	Boolean ok = false;
	trend ltrend;
	
	sm = hdGetMonthByDate(sYear, sMonth, &recnr, false);
	if ((sYear == eYear) && (sMonth == eMonth)) {
		em = sm;
	} else {
		em = hdGetMonthByDate(eYear, eMonth, &recnr, false);
	}
	
	if ((sm != NULL) && (em != NULL)) {
		DfitSlopeStart();
		sday = sDay;
		while (sm != em) {
			if (sm != NULL) {
				for (i = sday; i < sm->days; i++) {
					DfitSlopePoint(cvtTrend(sm->trendArray[i], (weightunit) sm->weightUnit, prefs.weightUnit));
				}
				ltrend = cvtTrend(sm->trendArray[sm->days - 1], (weightunit) sm->weightUnit, prefs.weightUnit);
				hdDisposeMonthUnpacked(sm);
			} else {
				//	Month within range not in database; fake it
				
				for (i = sday; i < DaysInMonth(sMonth + 1, sYear); i++) {
					DfitSlopePoint(ltrend);
				}
			}
			sday = 0;
			sMonth++;
			if (sMonth >= 12) {
				sYear++;
				sMonth = 0;
			}
			if ((sYear == eYear) && (sMonth == eMonth)) {
				sm = em;
			} else {
				sm = hdGetMonthByDate(sYear, sMonth, &recnr, false);
			}
		}
		for (i = sday; i <= eDay; i++) {
			DfitSlopePoint(cvtTrend(em->trendArray[i], (weightunit) em->weightUnit, prefs.weightUnit));
		}
		*slope = DfitSlopeEnd();
		ok = true;
	}
	
	if (sm != NULL) {
		if (sm == em) {
			em = NULL;
		}
		hdDisposeMonthUnpacked(sm);
	}
	if (em != NULL) {
		hdDisposeMonthUnpacked(em);
	}
	return ok;
}
#endif